# Simple Calculator Parser - Example Solution
# FGII, Fall Semester 2013, T. Klauser
#
# Description:
# This simple calculator parser evaluates any expression
# typed in, and outputs the result. To be used with python 3.x.x
#
# Example:
#   > 2 + 3 - 1
#   Result: 4
#
# See the modifications marked with (c), (d), and (bonus) below

def scanCommand():
    if scanAssignment():                # (d) include scanAssignment() into scanCommand()
        return

    ok, x = scanExpression()
    if ok:
        print("Result:", x)

    if scanPosition < len(scannedString):	# (bonus) Test that the entire string was scanned.
        print("Syntax error!")				#         Otherwise, print an error message
        print("The error is probably located around here:", scannedString[scanPosition:])

# (d) implement scanAssignment() to parse variable assignments
def scanAssignment():
    global scanPosition

    lastPosition = scanPosition
    ok, c = scanLetter()
    if ok and scanChar('='):
        ok, x = scanExpression()
        if ok:
            setVariable(c, x)
            return True

    scanPosition = lastPosition
    return False

def scanExpression():
    global scanPosition

    ok, x = scanTerm()              # (c) use scanTerm() instead of scanFactor()
    if not ok:
        return False, 0

    while True:
        lastPosition = scanPosition
        if scanChar('+'):
            ok, y = scanTerm()      # (c) use scanTerm() instead of scanFactor()
            if ok:
                x = x + y
            else:
                scanPosition = lastPosition
                break
        elif scanChar('-'):
            ok, y = scanTerm()      # (c) use scanTerm() instead of scanFactor()
            if ok:
                x = x - y
            else:
                scanPosition = lastPosition
                break
        else:
            break

    return True, x

# (c) implement scanTerm(), analogous to scanExpression()
def scanTerm():
    global scanPosition

    ok, x = scanFactor()
    if not ok:
        return False, 0

    while True:
        lastPosition = scanPosition
        if scanChar('x'):
            ok, y = scanFactor()
            if ok:
                x = x * y
            else:
                scanPosition = lastPosition
                break
        else:
            break

    return True, x

def scanFactor():
    global scanPosition                     # (c) add support for parentheses...
                                            #
    lastPosition = scanPosition             #
    if scanChar('('):                       #
        ok, x = scanExpression()            #
        if ok and scanChar(')'):            #
            return True, x                  #
        else:                               #
            scanPosition = lastPosition     #
            return False, 0                 # ...until here

    ok, x = scanNumber()
    if ok:
        return True, x

    ok, c = scanLetter()        # (d) include scanLetter() into scanFactor() to
    if ok:                      #     parse variable assignments
        x = getVariable(c)
        return True, x

    return False, 0

# Return whether the character 'char' could be scanned
def scanChar(char):
    skipWhitespace()
    ok, c = scanChars(char)
    return ok

# Return whether any sequence of digits could be scanned,
# together with its numerical value, as an integer number
def scanNumber():
    skipWhitespace()
    ok, c = scanDigit()
    if not ok: return False, 0
    s = ''
    while ok:
        s = s + c
        ok, c = scanDigit()
    return True, int(s)

# Return whether any digit could be scanned,
# together with the scanned digit
def scanDigit():
    return scanChars(['0', '1', '2', '3', '4', '5', '6', '7', '8', '9'])

# Return whether any letter (a, b or c) could be scanned,
# together with the scanned letter
def scanLetter():
    skipWhitespace()
    return scanChars(['a', 'b', 'c'])

# Skip whitespace characters
def skipWhitespace():
    while True:
        ok, c = scanChars([' '])
        if not ok:
            break;

# Return whether any character from the list 'chars' could be scanned,
# together with the scanned character
def scanChars(chars):
    global scannedString, scanPosition
    if scanPosition >= len(scannedString):
        return False, ''
    for c in chars:
        if scannedString[scanPosition] == c:
            scanPosition += 1
            return True, c
    else:
        return False, ''

# Initialize the sttring to be scanned
def setScannedString(string):
    global scannedString, scanPosition
    scannedString = string
    scanPosition = 0

# Set the content of variable 'var' to x
def setVariable(var, x):
    global memory
    memory[var] = x

# Get the content of variable 'var'
def getVariable(var):
    global memory
    # (optional) only access memory if variable is stored there to prevent getting exceptions
    if var in memory:
        return memory[var]
    else:
        print("unknown variable", var)
        return 0

memory = {}

# Main part of the program
print("(Type q to quit)")
while True:
    line = input("> ")      # Prompt user for input
    if line[:1] == 'q':
        break				# Stop if user types 'q'
    elif len(line) > 0:     # If user typed something
        setScannedString(line)
        scanCommand()
